#include "esp_system.h"
#include "sntp.h"
#include "time.h"
#include <WiFi.h>
#include <WiFiUdp.h>
#include <Preferences.h>
#include "scheduler.h"
#include "scheduler_prefs.h"


// Init const and global objects

// PREFRENCES
Preferences preferences;
#define RW_MODE false
#define RO_MODE true

// WIFI
const char* ssid       = "Wokwi-GUEST";
const char* password   = "";

// NTP
const char* ntpServer1 = "pool.ntp.org";
const char* ntpServer2 = "time.nist.gov";
const long  gmtOffset_sec = 2 * 3600;  // GMT + 2
const int   daylightOffset_sec = 0;

// TIMER
hw_timer_t*timer_main = NULL;
uint32_t timer_interval = 3000000;
uint16_t timer_interval_in_secs = timer_interval / 1000000;
;

// SCHEDULER
uint8_t current_hour = HOUR_DEFAULT;
uint8_t current_minute = MINUTE_DEFAULT;
uint8_t current_second = 0;
uint32_t current_ts = 0;
uint32_t next_event_ts = 0;
uint8_t scheduler_1[24] = {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};
uint8_t scheduler_2[24] = {1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
uint8_t* scheduler_list[] = {
  scheduler_1, scheduler_2
  };

// PINS
unsigned int PIN_1 = 2;
unsigned int PIN_2 = 4;
unsigned int pin_list[2] = { PIN_1, PIN_2 };

// MISC

char buffer[40];

// Local functions

// Time helper
uint32_t get_timestamp( uint8_t hour, uint8_t minute, uint8_t second){
  return ( 3600 * hour  + 60 * minute + second) % 86400;
}

bool event_keep_waiting( uint32_t current_ts, uint32_t event_ts ){

  // basic case
  // ts: 11h 41m 11s
  // ev: 11h 44m 0s
  // ts < ev => TRUE

  // ts: 11h 44m 11s
  // ev: 11h 44m 0s
  // ts > ev => FALSE

  // if time looped at midnight
  // ts: 23h 57m 21s
  // ev: 0h 1m 0s => 60
  // ts > ev => TRUE


  // Simple case
  if ( current_ts < event_ts ){
    return true;
  }
  if( (current_ts - event_ts) > 600 && event_ts < 300 ) {
    return true;
  }
  return false;
}


// NTP helper
void setLocalTime()
{
  struct tm timeinfo;
  if (!getLocalTime(&timeinfo)){
    Serial.println("No time available (yet)");
    return;
  }
  current_hour = timeinfo.tm_hour;
  current_minute = timeinfo.tm_min;
  current_second = timeinfo.tm_sec;
  Serial.println(&timeinfo, "%A, %B %d %Y %H:%M:%S");
}

// NTP Callback
void timeavailable(struct timeval *t)
{
  Serial.println("Got time adjustment from NTP!");

  setLocalTime();
}

// TIMER Callback
void IRAM_ATTR onTimer(){


  /*
  struct tm timeinfo;
  // Get the current time via NTP, or downgrade
  if (getLocalTime(&timeinfo)){
    Serial.println("onTimer::Got local time");
    current_hour = timeinfo.tm_hour;
    current_minute = timeinfo.tm_min;
    current_second = timeinfo.tm_sec;
  // If no NTP clock
  }else{
    */
    Serial.println("onTimer::NO NTP");
    current_second += timer_interval_in_secs;
    if(current_second >= 60) {
      current_second = 0;
      current_minute += 1;
      if(current_minute >= 60) {
        current_minute = 0;
        current_hour += 1;
        if(current_hour >=  24) {
          current_hour = 0;
        }
      }
    }
    /*
  }
  */
  // If not expected to run exit
  current_ts = get_timestamp( current_hour, current_minute, current_second );
  Serial.printf("onTimer::check event %d %d\n", current_ts, next_event_ts);
  if( event_keep_waiting( current_ts, next_event_ts  ) ){
    return;
  }

  // Set the next target
  next_event_ts = get_timestamp( current_hour, current_minute + EVENT_INC_MINUTE, current_second );
  Serial.println("Reconfigure the Relays!");

  // Run the relays reconfiguration
  for ( int i = 0; i < 2; i++ ) {
    // Get a pointer to the current array
    uint8_t* scheduler  = scheduler_list[i];
    unsigned int pin    = pin_list[i];
    // Set the expected status
    uint8_t min_fivers  = (current_minute / 5);
    uint8_t hour        = scheduler[current_hour];
    Serial.printf("Check %d >> %d\n", hour, min_fivers);
    if (( hour >> min_fivers ) & 1) {
      digitalWrite(pin, HIGH);
      sprintf (buffer, "Pin %d is up ", pin);
    } else {
      digitalWrite(pin, LOW);
      sprintf (buffer, "Pin %d is down ", pin);
    }
    Serial.println(buffer);    // Get the expected status
  }

}

void setup(){

  // HARDWARE
  Serial.begin(115200);
  delay(1000);
  Serial.println("Starting.");
  pinMode(PIN_1, OUTPUT);
  pinMode(PIN_2, OUTPUT);

  // PREFERENCES
  Serial.println("Setup::PREFERENCES");
  preferences.begin(PREF_NAMESPACE, RW_MODE);
  bool pref_init = preferences.isKey("nvsInit");
  if (pref_init == false) {
    preferences.putBytes(RELAY_1_SCHEDULE, scheduler_1_default, sizeof(scheduler_1_default));
    preferences.putBytes(RELAY_2_SCHEDULE, scheduler_2_default, sizeof(scheduler_2_default));
    /// ... more to come
  }
  /* Huh... This crashes wokwi
  preferences.getBytes(RELAY_1_SCHEDULE, scheduler_1, preferences.getBytesLength(RELAY_1_SCHEDULE));
  preferences.getBytes(RELAY_2_SCHEDULE, scheduler_2, preferences.getBytesLength(RELAY_2_SCHEDULE));
  */
  preferences.end();

  // WIFI
  Serial.println("Setup::WIFI");
  Serial.printf("Connecting to %s ", ssid);
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.print("Connect to IP Address: ");
  Serial.print("http://");
  Serial.println(WiFi.localIP());

  // TIMER
  Serial.println("Setup::Timer");
  timer_main = timerBegin(0, 80, true);
  timerAttachInterrupt(timer_main, &onTimer, true);
  timerAlarmWrite(timer_main, timer_interval, true);
  timerAlarmEnable(timer_main);

  // NTP
  Serial.println("Setup::NTP");
  sntp_set_time_sync_notification_cb(timeavailable);
  configTime(gmtOffset_sec, daylightOffset_sec, ntpServer1, ntpServer2);

  // SCHEDULER
  Serial.println("Setup::SCHEDULER");
  next_event_ts = get_timestamp( HOUR_DEFAULT, MINUTE_DEFAULT + EVENT_INC_MINUTE, 0 );


  // HTTP
  Serial.println("Setup::HTTP");
  // Todo
}
void loop(){

  Serial.println("Loop::Enter");

  delay(1000);

}
